///////////////////////////////////////////////////////////////////////////////
////  (c) Chaichuk AU (chaichuk@mail.ru), 								   ////
////  2004, Chelyabinsk, NTC Privodnaya Tehnika (momentum@chel.surnet.ru)  ////
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <time.h>

#include "../api.h"
#include "sfins_io.h"


//-----------------------------------------------------------------------------
// получить текущее время в виде строки (для сообщений об ошибках)
//-----------------------------------------------------------------------------
char* get_time_now ()
{
	time_t tod;
	static char out [100];
	
	time (&tod);
    strftime (out, sizeof (out), "%d-%m-%Y %T", localtime (&tod));	
    
    return out;
}

//-----------------------------------------------------------------------------
// открыть порт
//-----------------------------------------------------------------------------
int open_port (const char *port_name, int speed)
{
	int fd;
	struct termios termios_p;
	
	fd = open (port_name, O_RDWR);
	if (fd == -1) {
		perror ("Error opening port. ");
		return (-1);	
	}
	tcgetattr (fd, &termios_p);
	cfmakeraw (&termios_p);
	termios_p.c_cflag = CS7 | CSTOPB | PARENB | CREAD;
	termios_p.c_iflag = INPCK;
	termios_p.c_lflag = IEXTEN;
	termios_p.c_oflag = 0;
	termios_p.c_ispeed = speed;
	termios_p.c_ospeed = speed;
	termios_p.c_cc [VFWD] = '\r';
	tcsetattr (fd, TCSANOW, &termios_p);
	
	return fd;
}

//-----------------------------------------------------------------------------
// расчет контрольной суммы
//-----------------------------------------------------------------------------
unsigned short calc_fcs (const char * msg, int size)
{
	unsigned short fcs = 0;
	int i;
	
	for (i=0; i<size; i++)
		fcs ^= msg[i];	
	return fcs;
}

//-----------------------------------------------------------------------------
// отправить сообщение + контрольную сумму
//-----------------------------------------------------------------------------
int send_request (int port_fd, const char *msg)
{
	char buf[255];
	unsigned short fcs;
	
	fcs = calc_fcs (msg, strlen (msg));
	sprintf (buf, "%s%02X*\r", msg, fcs);
	
	if (write (port_fd, buf, strlen (buf)) == -1) {
		perror ("Error writing to port. ");
		return (-1);
	}
	return 0;
}	

//-----------------------------------------------------------------------------
// принять ответ
//-----------------------------------------------------------------------------
int read_response (int port_fd, char *msg, int len_expected)
{
	int ret;
	
	ret = readcond (port_fd, msg, len_expected, len_expected, 1, 1);
	if (ret == -1) {
		perror ("Error while reading response. ");
		return (-1);
	}
	if (len_expected > 0 && ret < len_expected) {
		fprintf (stderr, "%s - Error: The number of bytes read from port is less then expected\n", get_time_now ());
		return (-1);
	}
	return 0; 	
}

//-----------------------------------------------------------------------------
// запись данных по протоколу Hostlink
//-----------------------------------------------------------------------------
int write_CIO (int port_fd, int unit, int start_addr, short *buf, int n_words)
{
	char out_buf[255], in_buf[255];
	int i;
	int unit_resp, code_resp, fcs_resp;
	int count;
	
	sprintf (out_buf, "@%02dWR%04d", unit, start_addr);
	for (i=0; i<n_words; i++) {
		sprintf (out_buf, "%s%04X", out_buf, buf[i]);
	}

	if (send_request  (port_fd, out_buf) == -1) return (-1);
	if (read_response (port_fd, in_buf, 11) == -1) return (-1);
	
	count = sscanf (in_buf, "@%2dWR%2X%2X",	&unit_resp, &code_resp, &fcs_resp);
	if (count < 3) {
		fprintf (stderr, "%s - Write_CIO:Invalid response:\"%s\" \n", get_time_now (), in_buf);
		return (-1);
	}
	in_buf[7] = '\0';
	if (fcs_resp != calc_fcs (in_buf, strlen (in_buf))) {
		fprintf (stderr, "%s - Write_CIO:Invalid FCS (response) \n", get_time_now ());
		return (-1);
	}
	if (unit_resp != unit) {
		fprintf (stderr, "%s - Write_CIO:Invalid unit (response)\n", get_time_now ());
		return (-1);
	}			
	if (code_resp != 0x00) {
		switch (code_resp) {
			case 0x01: fprintf (stderr, "Write_CIO:Not executable in RUN mode\n");
				break;
			case 0x13: fprintf (stderr, "Write_CIO:FCS error\n");
				break;
			case 0x14: fprintf (stderr, "Write_CIO:Format error\n");
				break;
			case 0x15: fprintf (stderr, "Write_CIO:Entry number data error\n");
				break;
			case 0x18: fprintf (stderr, "Write_CIO:Frame lenght error\n");
				break;
			case 0x21: fprintf (stderr, "Write_CIO:Not executable due to CPU Unit CPU error\n");
				break;
			case 0xA3: fprintf (stderr, "Write_CIO:Aborted due to FCS error in transmit data\n");
				break;
			case 0xA4: fprintf (stderr, "Write_CIO:Aborted due to format error in transmit data\n");
				break;
			case 0xA5: fprintf (stderr, "Write_CIO:Aborted due to entry number data error in transmit data\n");
				break;
			case 0xA8: fprintf (stderr, "Write_CIO:Aborted due to frame lenght error in transmit data\n");
				break;
			default:   fprintf (stderr, "Write_CIO:Unknown end code\n");		
		}
		return (-1);
	}
	return 0;
}

//-----------------------------------------------------------------------------
// чтение данных по протоколу Hostlink
//-----------------------------------------------------------------------------
int read_CIO (int port_fd, int unit, int start_addr, short *buf, int n_words)
{
	char out_buf[255], in_buf[255];
	int len_expected, count;
	int unit_resp, code_resp, fcs_resp;
	int i;
	
	sprintf (out_buf, "@%02dRR%04d%04d", unit, start_addr, n_words);
	len_expected = 7 + 4*n_words + 4;
	
	if (send_request  (port_fd, out_buf) == -1) return (-1);
	if (read_response (port_fd, in_buf, len_expected) == -1) return (-1);

	count = sscanf (in_buf, "@%2dRR%2X", &unit_resp, &code_resp);	
	if (count < 2) {
		fprintf (stderr, "%s - Read_CIO: Invalid response (unit & code)\n", get_time_now ());
		return (-1);
	}
	sscanf (&in_buf [7 + 4*n_words], "%2X", &fcs_resp);
	in_buf [7 + 4*n_words] = '\0';
	if (fcs_resp != calc_fcs (in_buf, strlen (in_buf))) {
		fprintf (stderr, "%s - Read_CIO: Invalid FCS (response)\n", get_time_now ());
		return (-1);
	}
	if (code_resp != 0x00) {
		switch (code_resp) {
			case 0x13: fprintf (stderr, "Read_CIO: FCS error\n");
				break;
			case 0x14: fprintf (stderr, "Read_CIO: Format error\n");
				break;
			case 0x15: fprintf (stderr, "Read_CIO: Entry number data error\n");
				break;
			case 0x18: fprintf (stderr, "Read_CIO: Frame lenght error\n");
				break;
			case 0x21: fprintf (stderr, "Read_CIO: Not executable due to CPU Unit CPU error\n");
				break; 
			default:   fprintf (stderr, "Read_CIO: Unknown end code\n");
		}
		return (-1);
	}
	for (i=0; i<n_words; i++) {
		sscanf (&in_buf [7 + i*4], "%4hX", &buf[i]);
	}		
	return 0;
}

//----------------------------------------------------------------------------
// отправить запрос по протоколу FINS (локально)
//-----------------------------------------------------------------------------
int send_local_FINS (int port_fd, int unit, int SID, int command_code, 
				char *command_text, char *response_text, int response_len) 
{
	char out_buf [255], in_buf [1200];
	int len_expected, count;
	int unit_resp, SID_resp, code_resp, FINS_resp, fcs_resp;
	
	sprintf (out_buf, "@%02dFA0000000%02X%04X%s", unit, SID, command_code, command_text);
	len_expected = 23 + response_len + 4;
	
	if (send_request  (port_fd, out_buf) < 0) return (-1);
	if (read_response (port_fd, in_buf, len_expected) < 0) return (-1);

	count = sscanf (in_buf, "@%2dFA00400000%2X%4X%4X", &unit_resp, &SID_resp, 
					&code_resp, &FINS_resp);	
	if (count < 4) {
		fprintf (stderr, "%s - send_local_FINS: Invalid response\n", get_time_now ());
		return (-1);
	}
	if (SID_resp != SID) {
		fprintf (stderr, "%s - send_local_FINS: Invalid response SID\n", get_time_now ());
		return (-1);
	}
	if (code_resp != command_code) {
		fprintf (stderr, "%s - send_local_FINS: Invalid response command code\n", get_time_now ());
		return (-1);
	}
	if (FINS_resp != 0x0000) {
		fprintf (stderr, "%s - send_local_FINS: Error FINS response: %04X\n", get_time_now (), FINS_resp);
		return (-1);
	}
	
	sscanf (&in_buf [23 + response_len], "%2X", &fcs_resp);
	in_buf [23 + response_len] = '\0';
	if (fcs_resp != calc_fcs (in_buf, strlen (in_buf))) {
		fprintf (stderr, "%s - send_local_FINS: Invalid FCS (response)\n", get_time_now ());
		return (-1);
	}

	strcpy (response_text, &in_buf [23]);
	return 0;
}

//-----------------------------------------------------------------------------
// чтение данных по протоколу FINS (локально)
//-----------------------------------------------------------------------------
int FINS_memory_read (int port_fd, int unit, int memory_code, int start_addr, int n_items, short *buf)
{
	int command_code 	= 0x0101;
	int SID				= 0x00;
	int response_len;
	char command_text  [255];		
	char response_text [1200];
	int i;
	int return_code;
	
	if (!!(memory_code & 0x80)) response_len = n_items * 4;
	else response_len = n_items * 2;
	
	sprintf (command_text, "%02X%06X%04X", memory_code, start_addr, n_items);
	return_code = send_local_FINS (port_fd, unit, SID, command_code, command_text, response_text, response_len);
	if (return_code < 0) return return_code;
	
	if (!!(memory_code & 0x80)) {
		for (i = 0; i < n_items; i++) {
			sscanf (&response_text [i*4], "%4hX", &buf [i]); 	
		}
	} else {
			for (i = 0; i < n_items; i++) {
			sscanf (&response_text [i*2], "%2hX", &buf [i]); 	
		}
	}	
	return return_code;
}

//-----------------------------------------------------------------------------
// запись данных по протоколу FINS (локально)
//-----------------------------------------------------------------------------
int FINS_memory_write (int port_fd, int unit, int memory_code, int start_addr, int n_items, short *buf)
{
	int command_code 	= 0x0102;
	int SID				= 0x00;
	int response_len	= 0;
	char command_text  [255];		
	char response_text [1200];
	int i;

	sprintf (command_text, "%02X%06X%04X", memory_code, start_addr, n_items);
	if (!!(memory_code & 0x80)) {
		for (i = 0; i < n_items; i++) {
			sprintf (&command_text [12 + i*4], "%04hX", buf [i]); 	
		}
	} else {
		for (i = 0; i < n_items; i++) {
			sprintf (&command_text [12 + i*2], "%02hX", buf [i]); 	
		}
	}	
	return send_local_FINS (port_fd, unit, SID, command_code, command_text, response_text, response_len);
}